"""Stores time series metadata in a SQLite database."""

import hashlib
import itertools
import json
import os
import sqlite3
from dataclasses import dataclass
from typing import Any, Optional, Sequence
from uuid import UUID

from loguru import logger

from infrasys.exceptions import ISAlreadyAttached, ISOperationNotAllowed, ISNotStored
from infrasys import Component
from infrasys.serialization import (
    deserialize_value,
    serialize_value,
    SerializedTypeMetadata,
    TYPE_METADATA,
)
from infrasys.time_series_models import TimeSeriesMetadata
from infrasys.utils.sqlite import execute


class TimeSeriesMetadataStore:
    """Stores time series metadata in a SQLite database."""

    TABLE_NAME = "time_series_metadata"

    def __init__(self, con: sqlite3.Connection, initialize: bool = True):
        self._con = con
        if initialize:
            self._create_metadata_table()
        self._supports_sqlite_json = _does_sqlite_support_json()
        if not self._supports_sqlite_json:
            # This is true on Ubuntu 22.04, which is used by GitHub runners as of March 2024.
            # It is non-trivial to upgrade SQLite on those platforms.
            # There is code in this file to preserve behavior with less than optimal performance
            # in some cases. We can remove it when we're confident that users and runners have
            # newer SQLite versions.
            logger.debug(
                "SQLite version {} does not support JSON queries, and so time series queries may "
                "have degraded performance.",
                sqlite3.sqlite_version,
            )

    def _create_metadata_table(self):
        schema = [
            "id INTEGER PRIMARY KEY",
            "time_series_uuid TEXT",
            "time_series_type TEXT",
            "initial_time TEXT",
            "resolution TEXT",
            "variable_name TEXT",
            "component_uuid TEXT",
            "component_type TEXT",
            "user_attributes_hash TEXT",
            "metadata JSON",
        ]
        schema_text = ",".join(schema)
        cur = self._con.cursor()
        execute(cur, f"CREATE TABLE {self.TABLE_NAME}({schema_text})")
        self._create_indexes(cur)
        self._con.commit()
        logger.debug("Created in-memory time series metadata table")

    def _create_indexes(self, cur) -> None:
        # Index strategy:
        # 1. Optimize for these user queries with indexes:
        #    1a. all time series attached to one component
        #    1b. time series for one component + variable_name + type
        #    1c. time series for one component with all user attributes
        # 2. Optimize for checks at system.add_time_series. Use all fields and attribute hash.
        # 3. Optimize for returning all metadata for a time series UUID.
        execute(
            cur,
            f"CREATE INDEX by_c_vn_tst_hash ON {self.TABLE_NAME} "
            f"(component_uuid, variable_name, time_series_type, user_attributes_hash)",
        )
        execute(cur, f"CREATE INDEX by_ts_uuid ON {self.TABLE_NAME} (time_series_uuid)")

    def add(
        self,
        metadata: TimeSeriesMetadata,
        *components: Component,
    ) -> None:
        """Add metadata to the store.

        Raises
        ------
        ISAlreadyAttached
            Raised if the time series metadata already stored.
        """
        attribute_hash = _compute_user_attribute_hash(metadata.user_attributes)
        where_clause, params = self._make_where_clause(
            components,
            metadata.variable_name,
            metadata.type,
            attribute_hash=attribute_hash,
            **metadata.user_attributes,
        )
        cur = self._con.cursor()

        query = f"SELECT COUNT(*) FROM {self.TABLE_NAME} WHERE {where_clause}"
        res = execute(cur, query, params=params).fetchone()
        if res[0] > 0:
            msg = f"Time series with {metadata=} is already stored."
            raise ISAlreadyAttached(msg)

        rows = [
            (
                None,  # auto-assigned by sqlite
                str(metadata.time_series_uuid),
                metadata.type,
                str(metadata.initial_time),
                str(metadata.resolution),
                metadata.variable_name,
                str(component.uuid),
                component.__class__.__name__,
                attribute_hash,
                json.dumps(serialize_value(metadata)),
            )
            for component in components
        ]
        self._insert_rows(rows)

    def get_time_series_counts(self) -> "TimeSeriesCounts":
        """Return summary counts of components and time series."""
        query = f"""
            SELECT
                component_type
                ,time_series_type
                ,initial_time
                ,resolution
                ,count(*) AS count
            FROM {self.TABLE_NAME}
            GROUP BY
                component_type
                ,time_series_type
                ,initial_time
                ,resolution
            ORDER BY
                component_type
                ,time_series_type
                ,initial_time
                ,resolution
        """
        cur = self._con.cursor()
        rows = execute(cur, query).fetchall()
        time_series_type_count = {(x[0], x[1], x[2], x[3]): x[4] for x in rows}

        time_series_count = execute(
            cur, f"SELECT COUNT(DISTINCT time_series_uuid) from {self.TABLE_NAME}"
        ).fetchall()[0][0]

        return TimeSeriesCounts(
            time_series_count=time_series_count,
            time_series_type_count=time_series_type_count,
        )

    def get_metadata(
        self,
        component: Component,
        variable_name: Optional[str] = None,
        time_series_type: Optional[str] = None,
        **user_attributes,
    ) -> TimeSeriesMetadata:
        """Return the metadata matching the inputs.

        Raises
        ------
        ISOperationNotAllowed
            Raised if more than one metadata instance matches the inputs.
        """
        if variable_name is not None and time_series_type is not None:
            metadata = self._try_get_time_series_metadata_by_full_params(
                component, variable_name, time_series_type, **user_attributes
            )
            if metadata is not None:
                return metadata

        metadata_list = self.list_metadata(
            component,
            variable_name=variable_name,
            time_series_type=time_series_type,
            **user_attributes,
        )
        if not metadata_list:
            msg = "No time series matching the inputs is stored"
            raise ISNotStored(msg)

        if len(metadata_list) > 1:
            msg = f"Found more than metadata matching inputs: {len(metadata_list)}"
            raise ISOperationNotAllowed(msg)

        return metadata_list[0]

    def has_time_series(self, time_series_uuid: UUID) -> bool:
        """Return True if there is time series matching the UUID."""
        cur = self._con.cursor()
        query = f"SELECT COUNT(*) FROM {self.TABLE_NAME} WHERE time_series_uuid = ?"
        row = execute(cur, query, params=(str(time_series_uuid),)).fetchone()
        return row[0] > 0

    def has_time_series_metadata(
        self,
        component: Component,
        variable_name: Optional[str] = None,
        time_series_type: Optional[str] = None,
        **user_attributes: Any,
    ) -> bool:
        """Return True if there is time series metadata matching the inputs."""
        if (
            variable_name is not None
            and time_series_type is not None
            and self._try_has_time_series_metadata_by_full_params(
                component, variable_name, time_series_type, **user_attributes
            )
        ):
            return True

        if not self._supports_sqlite_json:
            return bool(
                self._list_metadata_no_sql_json(
                    component,
                    variable_name=variable_name,
                    time_series_type=time_series_type,
                    **user_attributes,
                )
            )

        where_clause, params = self._make_where_clause(
            (component,), variable_name, time_series_type, **user_attributes
        )
        query = f"SELECT COUNT(*) FROM {self.TABLE_NAME} WHERE {where_clause}"
        cur = self._con.cursor()
        res = execute(cur, query, params=params).fetchone()
        return res[0] > 0

    def list_existing_time_series(self, time_series_uuids: list[UUID]) -> set[UUID]:
        """Return the UUIDs that are present."""
        cur = self._con.cursor()
        params = tuple(str(x) for x in time_series_uuids)
        uuids = ",".join(itertools.repeat("?", len(time_series_uuids)))
        query = (
            f"SELECT time_series_uuid FROM {self.TABLE_NAME} WHERE time_series_uuid IN ({uuids})"
        )
        rows = execute(cur, query, params=params).fetchall()
        return {UUID(x[0]) for x in rows}

    def list_missing_time_series(self, time_series_uuids: list[UUID]) -> set[UUID]:
        """Return the UUIDs that are not present."""
        existing_uuids = set(self.list_existing_time_series(time_series_uuids))
        return set(time_series_uuids) - existing_uuids

    def list_metadata(
        self,
        *components: Component,
        variable_name: Optional[str] = None,
        time_series_type: Optional[str] = None,
        **user_attributes,
    ) -> list[TimeSeriesMetadata]:
        """Return a list of metadata that match the query."""
        if not self._supports_sqlite_json:
            return [
                x[1]
                for x in self._list_metadata_no_sql_json(
                    *components,
                    variable_name=variable_name,
                    time_series_type=time_series_type,
                    **user_attributes,
                )
            ]

        where_clause, params = self._make_where_clause(
            components, variable_name, time_series_type, **user_attributes
        )
        query = f"SELECT metadata FROM {self.TABLE_NAME} WHERE {where_clause}"
        cur = self._con.cursor()
        rows = execute(cur, query, params=params).fetchall()
        return [_deserialize_time_series_metadata(x[0]) for x in rows]

    def _list_metadata_no_sql_json(
        self,
        *components: Component,
        variable_name: Optional[str] = None,
        time_series_type: Optional[str] = None,
        **user_attributes,
    ) -> list[tuple[int, TimeSeriesMetadata]]:
        """Return a list of metadata that match the query.

        Returns
        -------
        list[tuple[int, TimeSeriesMetadata]]
            The first element of each tuple is the database id field that uniquely identifies the
            row.
        """
        where_clause, params = self._make_where_clause(components, variable_name, time_series_type)
        query = f"SELECT id, metadata FROM {self.TABLE_NAME} WHERE {where_clause}"
        cur = self._con.cursor()
        rows = execute(cur, query, params).fetchall()

        metadata_list = []
        for row in rows:
            metadata = _deserialize_time_series_metadata(row[1])
            if _do_attributes_match(metadata.user_attributes, **user_attributes):
                metadata_list.append((row[0], metadata))
        return metadata_list

    def list_rows(
        self,
        *components: Component,
        variable_name: Optional[str] = None,
        time_series_type: Optional[str] = None,
        columns=None,
        **user_attributes,
    ) -> list[tuple]:
        """Return a list of rows that match the query."""
        if not self._supports_sqlite_json and user_attributes:
            msg = (
                "list_rows is not supported with user_attributes with SQLite version "
                "{sqlite3.sqlite_version}"
            )
            raise ISOperationNotAllowed(msg)

        where_clause, params = self._make_where_clause(
            components, variable_name, time_series_type, **user_attributes
        )
        cols = "*" if columns is None else ",".join(columns)
        query = f"SELECT {cols} FROM {self.TABLE_NAME} WHERE {where_clause}"
        cur = self._con.cursor()
        rows = execute(cur, query, params=params).fetchall()
        return rows

    def remove(
        self,
        *components: Component,
        variable_name: str | None = None,
        time_series_type: Optional[str] = None,
        **user_attributes,
    ) -> list[UUID]:
        """Remove all matching rows and return the time series UUIDs."""
        cur = self._con.cursor()
        if not self._supports_sqlite_json:
            ts_uuids = set()
            ids = []
            for id_, metadata in self._list_metadata_no_sql_json(
                *components,
                variable_name=variable_name,
                time_series_type=time_series_type,
                **user_attributes,
            ):
                ts_uuids.add(metadata.time_series_uuid)
                ids.append(id_)
            params = [str(x) for x in ids]
            id_str = ",".join(itertools.repeat("?", len(ids)))
            query = f"DELETE FROM {self.TABLE_NAME} WHERE id IN ({id_str})"
            execute(cur, query, params=params)
            count_deleted = execute(cur, "SELECT changes()").fetchall()[0][0]
            if count_deleted != len(ids):
                msg = f"Bug: Unexpected length mismatch {len(ts_uuids)=} {count_deleted=}"
                raise Exception(msg)
            self._con.commit()
            return list(ts_uuids)

        where_clause, params = self._make_where_clause(
            components, variable_name, time_series_type, **user_attributes
        )
        query = f"SELECT time_series_uuid FROM {self.TABLE_NAME} WHERE {where_clause}"
        uuids = [UUID(x[0]) for x in execute(cur, query, params=params).fetchall()]

        query = f"DELETE FROM {self.TABLE_NAME} WHERE ({where_clause})"
        execute(cur, query, params=params)
        self._con.commit()
        count_deleted = execute(cur, "SELECT changes()").fetchall()[0][0]
        if len(uuids) != count_deleted:
            msg = f"Bug: Unexpected length mismatch: {len(uuids)=} {count_deleted=}"
            raise Exception(msg)
        return uuids

    def sql(self, query: str, params: Sequence[str] = ()) -> list[tuple]:
        """Run a SQL query on the time series metadata table."""
        cur = self._con.cursor()
        return execute(cur, query, params=params).fetchall()

    def _insert_rows(self, rows: list[tuple]) -> None:
        cur = self._con.cursor()
        placeholder = ",".join(["?"] * len(rows[0]))
        query = f"INSERT INTO {self.TABLE_NAME} VALUES({placeholder})"
        try:
            cur.executemany(query, rows)
        finally:
            self._con.commit()

    def _make_components_str(self, params: list[str], *components: Component) -> str:
        if not components:
            msg = "At least one component must be passed."
            raise ISOperationNotAllowed(msg)

        or_clause = "OR ".join((itertools.repeat("component_uuid = ? ", len(components))))

        for component in components:
            params.append(str(component.uuid))

        return f"({or_clause})"

    def _make_where_clause(
        self,
        components: tuple[Component, ...],
        variable_name: Optional[str],
        time_series_type: Optional[str],
        attribute_hash: Optional[str] = None,
        **user_attributes: str,
    ) -> tuple[str, list[str]]:
        params: list[str] = []
        component_str = self._make_components_str(params, *components)

        if variable_name is None:
            var_str = ""
        else:
            var_str = "AND variable_name = ?"
            params.append(variable_name)

        if time_series_type is None:
            ts_str = ""
        else:
            ts_str = "AND time_series_type = ?"
            params.append(time_series_type)

        if attribute_hash is None and user_attributes:
            _raise_if_unsupported_sql_operation()
            ua_hash_filter = _make_user_attribute_filter(user_attributes, params)
            ua_str = f"AND {ua_hash_filter}"
        else:
            ua_str = ""

        if attribute_hash:
            ua_hash_filter = _make_user_attribute_hash_filter(attribute_hash, params)
            ua_hash = f"AND {ua_hash_filter}"
        else:
            ua_hash = ""

        return f"({component_str} {var_str} {ts_str}) {ua_str} {ua_hash}", params

    def _try_time_series_metadata_by_full_params(
        self,
        component: Component,
        variable_name: str,
        time_series_type: str,
        column: str,
        **user_attributes: str,
    ) -> list[tuple] | None:
        assert variable_name is not None
        assert time_series_type is not None
        where_clause, params = self._make_where_clause(
            (component,),
            variable_name,
            time_series_type,
            attribute_hash=_compute_user_attribute_hash(user_attributes),
            **user_attributes,
        )
        query = f"SELECT {column} FROM {self.TABLE_NAME} WHERE {where_clause}"
        cur = self._con.cursor()
        rows = execute(cur, query, params=params).fetchall()
        if not rows:
            return None

        return rows

    def _try_get_time_series_metadata_by_full_params(
        self,
        component: Component,
        variable_name: str,
        time_series_type: str,
        **user_attributes: str,
    ) -> TimeSeriesMetadata | None:
        """Attempt to get the metadata by using all parameters.

        This will return the metadata if the user passes all user attributes that exist in the
        time series metadata. This is highly advantageous in cases where one component has a large
        number of time series and each metadata has user attributes. Otherwise, SQLite has to
        parse the JSON values.
        """
        rows = self._try_time_series_metadata_by_full_params(
            component,
            variable_name,
            time_series_type,
            "metadata",
            **user_attributes,
        )
        if rows is None:
            return rows

        if len(rows) > 1:
            msg = f"Found more than one metadata matching inputs: {len(rows)}"
            raise ISOperationNotAllowed(msg)

        return _deserialize_time_series_metadata(rows[0][0])

    def _try_has_time_series_metadata_by_full_params(
        self,
        component: Component,
        variable_name: str,
        time_series_type: str,
        **user_attributes: str,
    ) -> bool:
        """Attempt to check if the metadata is stored by using all parameters. Refer to
        _try_get_time_series_metadata_by_full_params for more information.
        """
        text = self._try_time_series_metadata_by_full_params(
            component,
            variable_name,
            time_series_type,
            "id",
            **user_attributes,
        )
        return text is not None

    def unique_uuids_by_type(self, time_series_type: str):
        query = (
            f"SELECT DISTINCT time_series_uuid from {self.TABLE_NAME} where time_series_type = ?"
        )
        params = (time_series_type,)
        uuid_strings = self.sql(query, params)
        return [UUID(ustr[0]) for ustr in uuid_strings]


@dataclass
class TimeSeriesCounts:
    """Summarizes the counts of time series by component type."""

    time_series_count: int
    # Keys are component_type, time_series_type, initial_time, resolution
    time_series_type_count: dict[tuple[str, str, str, str], int]


def _make_user_attribute_filter(user_attributes: dict[str, Any], params: list[str]) -> str:
    attrs = _make_user_attribute_dict(user_attributes)
    items = []
    for key, val in attrs.items():
        items.append(f"metadata->>'$.user_attributes.{key}' = ? ")
        params.append(val)
    return "AND ".join(items)


def _make_user_attribute_hash_filter(attribute_hash: str, params: list[str]) -> str:
    params.append(attribute_hash)
    return "user_attributes_hash = ?"


def _make_user_attribute_dict(user_attributes: dict[str, Any]) -> dict[str, Any]:
    return {k: user_attributes[k] for k in sorted(user_attributes)}


def _compute_user_attribute_hash(user_attributes: dict[str, Any]) -> str | None:
    if not user_attributes:
        return None

    attrs = _make_user_attribute_dict(user_attributes)
    return _compute_hash(bytes(json.dumps(attrs), encoding="utf-8"))


def _compute_hash(text: bytes) -> str:
    hash_obj = hashlib.sha256()
    hash_obj.update(text)
    return hash_obj.hexdigest()


def _deserialize_time_series_metadata(text: str) -> TimeSeriesMetadata:
    data = json.loads(text)
    type_metadata = SerializedTypeMetadata(**data.pop(TYPE_METADATA))
    metadata = deserialize_value(data, type_metadata.fields)
    return metadata


def _does_sqlite_support_json() -> bool:
    if "__INFRASYS_NON_JSON_SQLITE__" in os.environ:
        return False

    version = sqlite3.sqlite_version_info
    return not (version[0] == 3 and version[1] < 38)


def _raise_if_unsupported_sql_operation() -> None:
    if not _does_sqlite_support_json():
        msg = (
            "Operations that perform a JSON search with SQLite require version 3.38 or later: "
            f"{sqlite3.sqlite_version}. Please submit a bug report to the infrasys developers."
        )
        raise ISOperationNotAllowed(msg)


def _do_attributes_match(db_attributes: dict[str, Any], **user_attributes: str) -> bool:
    for key, val in user_attributes.items():
        if db_attributes.get(key) != val:
            return False
    return True



"""In-memory time series storage"""

from datetime import datetime
from pathlib import Path
import numpy as np
from numpy.typing import NDArray
from typing import Optional, TypeAlias
from uuid import UUID

from loguru import logger
from infrasys.arrow_storage import ArrowTimeSeriesStorage

from infrasys.exceptions import ISNotStored, ISOperationNotAllowed
from infrasys.time_series_models import (
    SingleTimeSeries,
    SingleTimeSeriesMetadata,
    TimeSeriesData,
    TimeSeriesMetadata,
)
from infrasys.time_series_storage_base import TimeSeriesStorageBase

DataStoreType: TypeAlias = NDArray


class InMemoryTimeSeriesStorage(TimeSeriesStorageBase):
    """Stores time series in memory."""

    def __init__(self) -> None:
        self._arrays: dict[UUID, DataStoreType] = {}  # Time series UUID, not metadata UUID

    def get_time_series_directory(self) -> None:
        return None

    def add_time_series(self, metadata: TimeSeriesMetadata, time_series: TimeSeriesData) -> None:
        if isinstance(time_series, SingleTimeSeries):
            if metadata.time_series_uuid not in self._arrays:
                self._arrays[metadata.time_series_uuid] = time_series.data_array
                logger.debug("Added {} to store", time_series.summary)
            else:
                logger.debug("{} was already stored", time_series.summary)

        else:
            msg = f"add_time_series not implemented for {type(time_series)}"
            raise NotImplementedError(msg)

    def add_raw_single_time_series(
        self, time_series_uuid: UUID, time_series_data: DataStoreType
    ) -> None:
        if time_series_uuid not in self._arrays:
            self._arrays[time_series_uuid] = time_series_data
            logger.debug("Added {} to store", time_series_uuid)
        else:
            logger.debug("{} was already stored", time_series_uuid)

    def get_time_series(
        self,
        metadata: TimeSeriesMetadata,
        start_time: datetime | None = None,
        length: int | None = None,
    ) -> TimeSeriesData:
        if isinstance(metadata, SingleTimeSeriesMetadata):
            return self._get_single_time_series(metadata, start_time, length)
        raise NotImplementedError(str(metadata.get_time_series_data_type()))

    def get_raw_single_time_series(self, time_series_uuid: UUID) -> NDArray:
        data_array = self._arrays[time_series_uuid]
        if not isinstance(data_array, np.ndarray):
            msg = f"Can't retrieve type: {type(data_array)} as single_time_series"
            raise ISOperationNotAllowed(msg)
        return data_array

    def remove_time_series(self, uuid: UUID) -> None:
        time_series = self._arrays.pop(uuid, None)
        if time_series is None:
            msg = f"No time series with {uuid} is stored"
            raise ISNotStored(msg)

    def serialize(self, dst: Path | str, _: Optional[Path | str] = None) -> None:
        base_directory = dst if isinstance(dst, Path) else Path(dst)
        storage = ArrowTimeSeriesStorage.create_with_permanent_directory(base_directory)
        for ts_uuid, ts in self._arrays.items():
            storage.add_raw_single_time_series(ts_uuid, ts)

    def _get_single_time_series(
        self,
        metadata: SingleTimeSeriesMetadata,
        start_time: datetime | None = None,
        length: int | None = None,
    ) -> SingleTimeSeries:
        ts_data = self._arrays.get(metadata.time_series_uuid)
        if ts_data is None:
            msg = f"No time series with {metadata.time_series_uuid} is stored"
            raise ISNotStored(msg)

        if start_time or length:
            index, length = metadata.get_range(start_time=start_time, length=length)
            ts_data = ts_data[index : index + length]

        if metadata.quantity_metadata is not None:
            ts_data = metadata.quantity_metadata.quantity_type(
                ts_data, metadata.quantity_metadata.units
            )
        return SingleTimeSeries(
            uuid=metadata.time_series_uuid,
            variable_name=metadata.variable_name,
            resolution=metadata.resolution,
            initial_time=start_time or metadata.initial_time,
            data=ts_data,
            normalization=metadata.normalization,
        )



import numpy as np

from infrasys.normalization import NormalizationMax, NormalizationByValue


def test_normalization_max():
    data = np.array([1.1, 2.2, 3.3, 4.4, 5.5])
    normalization = NormalizationMax()
    result = normalization.normalize_array(data)
    assert normalization.max_value == data[-1]
    for i, x in enumerate(result):
        assert x == data[i] / 5.5


def test_normalization_by_value():
    data = np.array([1.1, 2.2, 3.3, 4.4, 5.5])
    normalization = NormalizationByValue(value=data[-1])
    result = normalization.normalize_array(data)
    for i, x in enumerate(result):
        assert x == data[i] / 5.5



"""Base models for the package"""

import abc
from typing import Any
from uuid import UUID, uuid4

from loguru import logger
from pydantic import BaseModel, ConfigDict, Field, field_serializer


def make_model_config(**kwargs: Any) -> ConfigDict:
    """Return a Pydantic config"""
    return ConfigDict(
        str_strip_whitespace=True,
        validate_assignment=True,
        validate_default=True,
        extra="forbid",
        use_enum_values=False,
        arbitrary_types_allowed=True,
        populate_by_name=True,
        **kwargs,  # type: ignore
    )


class InfraSysBaseModel(BaseModel):
    """Base class for all Infrastructure Systems models"""

    model_config = make_model_config()


class InfraSysBaseModelWithIdentifers(InfraSysBaseModel, abc.ABC):
    """Base class for all Infrastructure Systems types with UUIDs"""

    uuid: UUID = Field(default_factory=uuid4, repr=False)

    @field_serializer("uuid")
    def _serialize_uuid(self, _) -> str:
        return str(self.uuid)

    def assign_new_uuid(self):
        """Generate a new UUID."""
        self.uuid = uuid4()
        logger.debug("Assigned new UUID for %s: %s", self.label, self.uuid)

    @classmethod
    def example(cls) -> "InfraSysBaseModelWithIdentifers":
        """Return an example instance of the model.

        Raises
        ------
        NotImplementedError
            Raised if the model does not implement this method.
        """
        msg = f"{cls.__name__} does not implement example()"
        raise NotImplementedError(msg)

    @property
    def label(self) -> str:
        """Provides a description of an instance."""
        class_name = self.__class__.__name__
        name = getattr(self, "name", "") or str(self.uuid)
        return make_label(class_name, name)


def make_label(class_name: str, name: str) -> str:
    """Make a string label of an instance."""
    return f"{class_name}.{name}"


def get_class_and_name_from_label(label: str) -> tuple[str, str | UUID]:
    """Return the class and name from a label.
    If the name is a stringified UUID, it will be converted to a UUID.
    """
    class_name, name = label.split(".", maxsplit=1)
    name_or_uuid: str | UUID = name
    try:
        name_or_uuid = UUID(name)
    except ValueError:
        pass

    return class_name, name_or_uuid



# ruff: noqa: F401
import importlib.metadata as metadata
from loguru import logger

logger.disable("infrasys")

__version__ = metadata.metadata("infrasys")["Version"]

from .component import Component
from .base_quantity import BaseQuantity
from .location import Location
from .normalization import NormalizationModel
from .system import System
from .time_series_models import SingleTimeSeries



"""Parquet time series storage"""

from datetime import datetime
from uuid import UUID

from infrasys.time_series_models import TimeSeriesData, TimeSeriesMetadata
from infrasys.time_series_storage_base import TimeSeriesStorageBase


class ParquetTimeSeriesStorage(TimeSeriesStorageBase):
    """Stores time series in Parquet files."""

    def add_time_series(self, metadata: TimeSeriesMetadata, time_series: TimeSeriesData) -> None:
        ...

    def get_time_series(
        self,
        metadata: TimeSeriesMetadata,
        start_time: datetime | None = None,
        length: int | None = None,
    ) -> TimeSeriesData:
        msg = "ParquetTimeSeriesStorage.get_time_series"
        raise NotImplementedError(msg)

    def remove_time_series(self, uuid: UUID) -> None:
        ...



"""Defines classes for value curves using cost functions"""

from typing_extensions import Annotated
from infrasys.exceptions import ISOperationNotAllowed
from infrasys.function_data import (
    LinearFunctionData,
    QuadraticFunctionData,
    PiecewiseLinearData,
    PiecewiseStepData,
    XYCoords,
    running_sum,
)
from pydantic import Field
import numpy as np

from infrasys.models import InfraSysBaseModelWithIdentifers


class ValueCurve(InfraSysBaseModelWithIdentifers):
    input_at_zero: Annotated[
        float | None,
        Field(
            description="Optional, an explicit representation of the input value at zero output."
        ),
    ] = None


class InputOutputCurve(ValueCurve):
    """Input-output curve relating production quality to cost.

    An input-output curve, directly relating the production quantity to the cost:

    .. math:: y = f(x).

    Can be used, for instance, in the representation of a Cost Curve where :math:`x` is MW and
    :math:`y` is currency/hr, or in the representation of a Fuel Curve where :math:`x` is MW and
    :math:`y` is fuel/hr.
    """

    function_data: Annotated[
        LinearFunctionData | QuadraticFunctionData | PiecewiseLinearData,
        Field(description="The underlying `FunctionData` representation of this `ValueCurve`"),
    ]


class IncrementalCurve(ValueCurve):
    """Incremental/marginal curve to relate production quantity to cost derivative.

    An incremental (or 'marginal') curve, relating the production quantity to the derivative of
    cost:

    ..math:: y = f'(x).

    Can be used, for instance, in the representation of a Cost Curve
    where :math:`x` is MW and :math:`y` is currency/MWh, or in the representation of a Fuel Curve
    where :math:`x` is MW and :math:`y` is fuel/MWh.
    """

    function_data: Annotated[
        LinearFunctionData | PiecewiseStepData,
        Field(description="The underlying `FunctionData` representation of this `ValueCurve`"),
    ]
    initial_input: Annotated[
        float | None,
        Field(
            description="The value of f(x) at the least x for which the function is defined, or \
                the origin for functions with no left endpoint, used for conversion to `InputOutputCurve`"
        ),
    ]

    def to_input_output(self) -> InputOutputCurve:
        """Function to convert IncrementalCurve to InputOutputCurve

        Function takes an IncrementalCurve and converts it to a corresponding
        InputOutputCurve depending on the type of function_data. If the IncrementalCurve
        uses LinearFunctionData, the new InputOutputCurve is created linear or quadratic data
        that correspond to the integral of the original linear function. If the input uses
        PiecewiseStepData, the slopes of each segment are used to calculate the corresponding
        y values for each x value and used to construct PiecewiseLinearData for the InputOutputCurve.

        Parameters
        ----------
        data : IncrementalCurve
            Original IncrementalCurve for conversion.

        Returns
        ----------
        InputOutputCurve
            InputOutputCurve using either QuadraticFunctionData or PiecewiseStepData.
        """
        match self.function_data:
            case LinearFunctionData():
                p = self.function_data.proportional_term
                m = self.function_data.constant_term

                c = self.initial_input
                if c is None:
                    msg = "Cannot convert `IncrementalCurve` with undefined `initial_input`"
                    raise ISOperationNotAllowed(msg)

                if p == 0:
                    return InputOutputCurve(
                        function_data=LinearFunctionData(proportional_term=m, constant_term=c)
                    )
                else:
                    return InputOutputCurve(
                        function_data=QuadraticFunctionData(
                            quadratic_term=p / 2, proportional_term=m, constant_term=c
                        ),
                        input_at_zero=self.input_at_zero,
                    )
            case PiecewiseStepData():
                c = self.initial_input
                if c is None:
                    msg = "Cannot convert `IncrementalCurve` with undefined `initial_input`"
                    raise ISOperationNotAllowed(msg)

                points = running_sum(self.function_data)

                return InputOutputCurve(
                    function_data=PiecewiseLinearData(
                        points=[XYCoords(p.x, p.y + c) for p in points]
                    ),
                    input_at_zero=self.input_at_zero,
                )


class AverageRateCurve(ValueCurve):
    """Average rate curve relating production quality to average cost rate.

    An average rate curve, relating the production quantity to the average cost rate from the
    origin:

    .. math:: y = f(x)/x.

    Can be used, for instance, in the representation of a
    Cost Curve where :math:`x` is MW and :math:`y` is currency/MWh, or in the representation of a
    Fuel Curve where :math:`x` is MW and :math:`y` is fuel/MWh. Typically calculated by dividing
    absolute values of cost rate or fuel input rate by absolute values of electric power.
    """

    function_data: Annotated[
        LinearFunctionData | PiecewiseStepData,
        Field(
            description="The underlying `FunctionData` representation of this `ValueCurve`, or \
                only the oblique asymptote when using `LinearFunctionData`"
        ),
    ]
    initial_input: Annotated[
        float | None,
        Field(
            description="The value of f(x) at the least x for which the function is defined, or \
                the origin for functions with no left endpoint, used for conversion to `InputOutputCurve`"
        ),
    ]

    def to_input_output(self) -> InputOutputCurve:
        """Function to convert IncrementalCurve to InputOutputCurve

        Function takes an AverageRateCurve and converts it to a corresponding
        InputOutputCurve depending on the type of function_data. If the AverageRateCurve
        uses LinearFunctionData, the new InputOutputCurve is created with either linear or quadratic
        function data, depending on if the original function data is constant or linear. If the
        input uses PiecewiseStepData, new y-values are calculated for each x value such that `f(x) = x*y`
        and used to construct PiecewiseLinearData for the InputOutputCurve.

        Parameters
        ----------
        data : AverageRateCurve
            Original AverageRateCurve for conversion.

        Returns
        ----------
        InputOutputCurve
            InputOutputCurve using either QuadraticFunctionData or PiecewiseStepData.
        """
        match self.function_data:
            case LinearFunctionData():
                p = self.function_data.proportional_term
                m = self.function_data.constant_term

                c = self.initial_input
                if c is None:
                    msg = "Cannot convert `AverageRateCurve` with undefined `initial_input`"
                    raise ISOperationNotAllowed(msg)

                if p == 0:
                    return InputOutputCurve(
                        function_data=LinearFunctionData(proportional_term=m, constant_term=c)
                    )
                else:
                    return InputOutputCurve(
                        function_data=QuadraticFunctionData(
                            # issue 53
                            quadratic_term=p,
                            proportional_term=m,
                            constant_term=c,  # type: ignore
                        ),
                        input_at_zero=self.input_at_zero,
                    )
            case PiecewiseStepData():
                c = self.initial_input
                if c is None:
                    msg = "Cannot convert `AverageRateCurve` with undefined `initial_input`"
                    raise ISOperationNotAllowed(msg)

                xs = self.function_data.x_coords
                ys = np.multiply(xs[1:], self.function_data.y_coords).tolist()
                ys.insert(0, c)

                return InputOutputCurve(
                    function_data=PiecewiseLinearData(
                        points=[XYCoords(x, y) for x, y in zip(xs, ys)]
                    ),
                    input_at_zero=self.input_at_zero,
                )
            case _:
                msg = "Function is not valid for the type of data provided."
                raise ISOperationNotAllowed(msg)


def LinearCurve(proportional_term: float = 0.0, constant_term: float = 0.0) -> InputOutputCurve:
    """Creates a linear curve using the given proportional and constant terms.

    Returns an instance of `InputOutputCurve` with the specified linear function parameters.

    If no arguments are provided, both the `proportional_term` and `constant_term` default to 0.

    Parameters
    ----------
    proportional_term : float, optional
        The slope of the linear curve. Defaults to 0.0.
    constant_term : float, optional
        The y-intercept of the linear curve. Defaults to 0.0.

    Returns
    -------
    InputOutputCurve
        An instance of `InputOutputCurve` with a `LinearFunctionData` object based on the given parameters.

    Examples
    --------
    >>> LinearCurve()
    InputOutputCurve(function_data=LinearFunctionData(proportional_term=0.0, constant_term=0.0))

    >>> LinearCurve(10)
    InputOutputCurve(function_data=LinearFunctionData(proportional_term=10.0, constant_term=0.0))

    >>> LinearCurve(10, 20)
    InputOutputCurve(function_data=LinearFunctionData(proportional_term=10.0, constant_term=20.0))

    >>> LinearCurve(proportional_term=5.0, constant_term=15.0)
    InputOutputCurve(function_data=LinearFunctionData(proportional_term=5.0, constant_term=15.0))
    """
    return InputOutputCurve(
        function_data=LinearFunctionData(
            proportional_term=proportional_term, constant_term=constant_term
        )
    )



from infrasys.function_data import (
    LinearFunctionData,
    QuadraticFunctionData,
    PiecewiseStepData,
)
from infrasys.value_curves import (
    InputOutputCurve,
    IncrementalCurve,
    AverageRateCurve,
    LinearCurve,
)
from infrasys import Component
from infrasys.exceptions import ISOperationNotAllowed
from .models.simple_system import SimpleSystem
import pytest


class ValueCurveComponent(Component):
    value_curve: InputOutputCurve


def test_input_output_curve():
    curve = InputOutputCurve(
        function_data=LinearFunctionData(proportional_term=1.0, constant_term=1.0)
    )

    assert isinstance(curve, InputOutputCurve)
    assert isinstance(curve.function_data, LinearFunctionData)


def test_incremental_curve():
    curve = IncrementalCurve(
        function_data=LinearFunctionData(proportional_term=1.0, constant_term=1.0),
        initial_input=1.0,
    )

    assert isinstance(curve, IncrementalCurve)
    assert isinstance(curve.function_data, LinearFunctionData)


def test_average_rate_curve():
    curve = AverageRateCurve(
        function_data=LinearFunctionData(proportional_term=1.0, constant_term=1.0),
        initial_input=1.0,
    )

    assert isinstance(curve, AverageRateCurve)
    assert isinstance(curve.function_data, LinearFunctionData)


def test_linear_curve():
    linear = LinearCurve()
    assert isinstance(linear, InputOutputCurve)
    assert getattr(linear, "function_data")
    assert getattr(linear.function_data, "proportional_term") == 0.0
    assert getattr(linear.function_data, "constant_term") == 0.0

    m = 15.0
    linear = LinearCurve(m)
    assert isinstance(linear, InputOutputCurve)
    assert getattr(linear, "function_data")
    assert getattr(linear.function_data, "proportional_term") == m
    assert getattr(linear.function_data, "constant_term") == 0.0

    m = 10.0
    b = 30.4
    linear = LinearCurve(m, b)
    assert isinstance(linear, InputOutputCurve)
    assert getattr(linear, "function_data")
    assert getattr(linear.function_data, "proportional_term") == m
    assert getattr(linear.function_data, "constant_term") == b


def test_average_rate_conversion():
    # Linear function data
    curve = AverageRateCurve(
        function_data=LinearFunctionData(proportional_term=1.0, constant_term=2.0),
        initial_input=None,
    )
    with pytest.raises(ISOperationNotAllowed):
        curve.to_input_output()

    curve.initial_input = 0.0
    new_curve = curve.to_input_output()
    assert isinstance(new_curve, InputOutputCurve)
    assert isinstance(new_curve.function_data, QuadraticFunctionData)
    assert new_curve.function_data.quadratic_term == 1.0

    assert isinstance(curve.function_data, LinearFunctionData)
    curve.function_data.proportional_term = 0.0
    new_curve = curve.to_input_output()
    assert isinstance(new_curve, InputOutputCurve)
    assert isinstance(new_curve.function_data, LinearFunctionData)
    assert new_curve.function_data.proportional_term == 2.0

    # Piecewise step data
    data = PiecewiseStepData(x_coords=[1.0, 3.0, 5.0], y_coords=[2.0, 6.0])
    curve = AverageRateCurve(function_data=data, initial_input=None)
    with pytest.raises(ISOperationNotAllowed):
        curve.to_input_output()

    curve.initial_input = 0.0
    new_curve = curve.to_input_output()
    assert isinstance(new_curve, InputOutputCurve)


def test_incremental_conversion():
    # Linear function data
    curve = IncrementalCurve(
        function_data=LinearFunctionData(proportional_term=1.0, constant_term=1.0),
        initial_input=None,
    )
    with pytest.raises(ISOperationNotAllowed):
        curve.to_input_output()
    curve.initial_input = 0.0
    new_curve = curve.to_input_output()
    assert isinstance(new_curve, InputOutputCurve)
    assert isinstance(new_curve.function_data, QuadraticFunctionData)
    assert new_curve.function_data.quadratic_term == 0.5

    assert isinstance(curve.function_data, LinearFunctionData)
    curve.function_data.proportional_term = 0.0
    new_curve = curve.to_input_output()
    assert isinstance(new_curve, InputOutputCurve)
    assert isinstance(new_curve.function_data, LinearFunctionData)
    assert new_curve.function_data.proportional_term == 1.0

    # Piecewise step data
    data = PiecewiseStepData(x_coords=[1.0, 3.0, 5.0], y_coords=[2.0, 6.0])
    curve = IncrementalCurve(function_data=data, initial_input=None)
    with pytest.raises(ISOperationNotAllowed):
        curve.to_input_output()
    curve.initial_input = 0.0
    new_curve = curve.to_input_output()
    assert isinstance(new_curve, InputOutputCurve)


def test_value_curve_custom_serialization():
    component = ValueCurveComponent(
        name="test",
        value_curve=InputOutputCurve(
            function_data=LinearFunctionData(proportional_term=1.0, constant_term=2.0)
        ),
    )

    model_dump = component.model_dump(mode="json")
    assert model_dump["value_curve"]["function_data"]["proportional_term"] == 1.0

    model_dump = component.model_dump(context={"magnitude_only": True})
    assert model_dump["value_curve"]["function_data"]["proportional_term"] == 1.0

    model_dump = component.model_dump(mode="json", context={"magnitude_only": True})
    assert model_dump["value_curve"]["function_data"]["proportional_term"] == 1.0


def test_value_curve_serialization(tmp_path):
    system = SimpleSystem(auto_add_composed_components=True)

    v1 = ValueCurveComponent(
        name="test",
        value_curve=InputOutputCurve(
            function_data=LinearFunctionData(proportional_term=1.0, constant_term=2.0)
        ),
    )
    system.add_component(v1)
    filename = tmp_path / "value_curve.json"

    system.to_json(filename, overwrite=True)
    system2 = SimpleSystem.from_json(filename)

    assert system2 is not None

    v2 = system2.get_component(ValueCurveComponent, "test")

    assert v2 is not None
    assert isinstance(v1.value_curve.function_data, LinearFunctionData)
    assert (
        v1.value_curve.function_data.proportional_term
        == v2.value_curve.function_data.proportional_term
    )
    assert v1.value_curve.function_data.constant_term == v2.value_curve.function_data.constant_term



"""Manages time series arrays"""

import sqlite3
from datetime import datetime
from pathlib import Path
from typing import Any, Optional, Type

from loguru import logger

from infrasys.arrow_storage import ArrowTimeSeriesStorage
from infrasys import Component
from infrasys.exceptions import ISInvalidParameter, ISOperationNotAllowed
from infrasys.in_memory_time_series_storage import InMemoryTimeSeriesStorage
from infrasys.time_series_metadata_store import TimeSeriesMetadataStore
from infrasys.time_series_models import (
    SingleTimeSeries,
    TimeSeriesData,
    TimeSeriesMetadata,
)
from infrasys.time_series_storage_base import TimeSeriesStorageBase

TIME_SERIES_KWARGS = {
    "time_series_in_memory": False,
    "time_series_read_only": False,
    "time_series_directory": None,
}


def _process_time_series_kwarg(key: str, **kwargs: Any) -> Any:
    return kwargs.get(key, TIME_SERIES_KWARGS[key])


class TimeSeriesManager:
    """Manages time series for a system."""

    def __init__(
        self,
        con: sqlite3.Connection,
        storage: Optional[TimeSeriesStorageBase] = None,
        initialize: bool = True,
        **kwargs,
    ) -> None:
        self._read_only = _process_time_series_kwarg("time_series_read_only", **kwargs)
        self._storage = storage or self.create_new_storage(**kwargs)
        self._metadata_store = TimeSeriesMetadataStore(con, initialize=initialize)

        # TODO: create parsing mechanism? CSV, CSV + JSON

    @staticmethod
    def create_new_storage(permanent: bool = False, **kwargs):
        base_directory: Path | None = _process_time_series_kwarg("time_series_directory", **kwargs)

        if _process_time_series_kwarg("time_series_in_memory", **kwargs):
            return InMemoryTimeSeriesStorage()
        else:
            if permanent:
                if base_directory is None:
                    msg = "Can't convert to perminant storage without a base directory"
                    raise ISInvalidParameter(msg)
                return ArrowTimeSeriesStorage.create_with_permanent_directory(
                    directory=base_directory
                )

            return ArrowTimeSeriesStorage.create_with_temp_directory(base_directory=base_directory)

    @property
    def metadata_store(self) -> TimeSeriesMetadataStore:
        """Return the time series metadata store."""
        return self._metadata_store

    @property
    def storage(self) -> TimeSeriesStorageBase:
        """Return the time series storage object."""
        return self._storage

    def add(
        self,
        time_series: TimeSeriesData,
        *components: Component,
        **user_attributes: Any,
    ) -> None:
        """Store a time series array for one or more components.

        Parameters
        ----------
        time_series : TimeSeriesData
            Time series data to store.
        components : Component
            Add the time series to all of these components.
        user_attributes : Any
            Key/value pairs to store with the time series data. Must be JSON-serializable.

        Raises
        ------
        ISAlreadyAttached
            Raised if the variable name and user attributes match any time series already
            attached to one of the components.
        ISOperationNotAllowed
            Raised if the manager was created in read-only mode.
        """
        self._handle_read_only()
        if not components:
            msg = "add_time_series requires at least one component"
            raise ISOperationNotAllowed(msg)

        ts_type = type(time_series)
        if not issubclass(ts_type, TimeSeriesData):
            msg = f"The first argument must be an instance of TimeSeriesData: {ts_type}"
            raise ValueError(msg)
        metadata_type = ts_type.get_time_series_metadata_type()
        metadata = metadata_type.from_data(time_series, **user_attributes)

        if not self._metadata_store.has_time_series(time_series.uuid):
            self._storage.add_time_series(metadata, time_series)

        self._metadata_store.add(metadata, *components)

    def get(
        self,
        component: Component,
        variable_name: str | None = None,
        time_series_type: Type[TimeSeriesData] = SingleTimeSeries,
        start_time: datetime | None = None,
        length: int | None = None,
        **user_attributes,
    ) -> TimeSeriesData:
        """Return a time series array.

        Raises
        ------
        ISNotStored
            Raised if no time series matches the inputs.
            Raised if the inputs match more than one time series.
        ISOperationNotAllowed
            Raised if the inputs match more than one time series.

        See Also
        --------
        list_time_series
        """
        metadata = self._metadata_store.get_metadata(
            component,
            variable_name=variable_name,
            time_series_type=time_series_type.__name__,
            **user_attributes,
        )
        return self._get_by_metadata(metadata, start_time=start_time, length=length)

    def has_time_series(
        self,
        component: Component,
        variable_name: str | None = None,
        time_series_type: Type[TimeSeriesData] = SingleTimeSeries,
        **user_attributes,
    ) -> bool:
        """Return True if the component has time series matching the inputs."""
        return self._metadata_store.has_time_series_metadata(
            component,
            variable_name=variable_name,
            time_series_type=time_series_type.__name__,
            **user_attributes,
        )

    def list_time_series(
        self,
        component: Component,
        variable_name: str | None = None,
        time_series_type: Type[TimeSeriesData] = SingleTimeSeries,
        start_time: datetime | None = None,
        length: int | None = None,
        **user_attributes: Any,
    ) -> list[TimeSeriesData]:
        """Return all time series that match the inputs."""
        metadata = self.list_time_series_metadata(
            component,
            variable_name=variable_name,
            time_series_type=time_series_type,
            **user_attributes,
        )
        return [self._get_by_metadata(x, start_time=start_time, length=length) for x in metadata]

    def list_time_series_metadata(
        self,
        component: Component,
        variable_name: str | None = None,
        time_series_type: Type[TimeSeriesData] = SingleTimeSeries,
        **user_attributes: Any,
    ) -> list[TimeSeriesMetadata]:
        """Return all time series metadata that match the inputs."""
        return self._metadata_store.list_metadata(
            component,
            variable_name=variable_name,
            time_series_type=time_series_type.__name__,
            **user_attributes,
        )

    def remove(
        self,
        *components: Component,
        variable_name: str | None = None,
        time_series_type: Type[TimeSeriesData] = SingleTimeSeries,
        **user_attributes: Any,
    ):
        """Remove all time series arrays matching the inputs.

        Raises
        ------
        ISNotStored
            Raised if no time series match the inputs.
        ISOperationNotAllowed
            Raised if the manager was created in read-only mode.
        """
        self._handle_read_only()
        time_series_uuids = self._metadata_store.remove(
            *components,
            variable_name=variable_name,
            time_series_type=time_series_type.__name__,
            **user_attributes,
        )
        missing_uuids = self._metadata_store.list_missing_time_series(time_series_uuids)
        for uuid in missing_uuids:
            self._storage.remove_time_series(uuid)
            logger.info("Removed time series {}.{}", time_series_type, variable_name)

    def copy(
        self,
        dst: Component,
        src: Component,
        name_mapping: dict[str, str] | None = None,
    ) -> None:
        """Copy all time series from src to dst.

        Parameters
        ----------
        dst : Component
            Destination component
        src : Component
            Source component
        name_mapping : dict[str, str]
            Optionally map src names to different dst names.
            If provided and src has a time_series with a name not present in name_mapping, that
            time_series will not copied. If name_mapping is nothing then all time_series will be
            copied with src's names.

        Notes
        -----
        name_mapping is currently not implemented.
        """
        self._handle_read_only()
        raise NotImplementedError

    def _get_by_metadata(
        self,
        metadata: TimeSeriesMetadata,
        start_time: datetime | None = None,
        length: int | None = None,
    ) -> TimeSeriesData:
        return self._storage.get_time_series(
            metadata,
            start_time=start_time,
            length=length,
        )

    def serialize(self, dst: Path | str, src: Optional[Path | str] = None) -> None:
        """Serialize the time series data to dst."""
        self._storage.serialize(dst, src)

    @classmethod
    def deserialize(
        cls,
        con: sqlite3.Connection,
        data: dict[str, Any],
        parent_dir: Path | str,
        **kwargs: Any,
    ) -> "TimeSeriesManager":
        """Deserialize the class. Must also call add_reference_counts after deserializing
        components.
        """
        time_series_dir = Path(parent_dir) / data["directory"]

        if _process_time_series_kwarg("time_series_read_only", **kwargs):
            storage = ArrowTimeSeriesStorage.create_with_permanent_directory(time_series_dir)
        else:
            storage = ArrowTimeSeriesStorage.create_with_temp_directory()
            storage.serialize(src=time_series_dir, dst=storage.get_time_series_directory())

        cls_instance = cls(con, storage=storage, initialize=False, **kwargs)

        if _process_time_series_kwarg("time_series_in_memory", **kwargs):
            cls_instance.convert_storage(**kwargs)

        return cls_instance

    def _handle_read_only(self) -> None:
        if self._read_only:
            msg = "Cannot modify time series in read-only mode."
            raise ISOperationNotAllowed(msg)

    def convert_storage(self, **kwargs) -> None:
        """
        Create a new storage instance and copy all time series from the current to new storage
        """
        new_storage = self.create_new_storage(**kwargs)
        for time_series_uuid in self.metadata_store.unique_uuids_by_type(
            SingleTimeSeries.__name__
        ):
            new_storage.add_raw_single_time_series(
                time_series_uuid, self._storage.get_raw_single_time_series(time_series_uuid)
            )

        self._storage = new_storage
        return None



from infrasys.cost_curves import CostCurve, FuelCurve
from infrasys.function_data import LinearFunctionData
from infrasys.value_curves import InputOutputCurve
from infrasys import Component
from .models.simple_system import SimpleSystem


class CurveComponent(Component):
    cost_curve: CostCurve


def test_cost_curve():
    # Cost curve
    cost_curve = CostCurve(
        value_curve=InputOutputCurve(
            function_data=LinearFunctionData(proportional_term=1.0, constant_term=2.0)
        ),
        vom_cost=InputOutputCurve(
            function_data=LinearFunctionData(proportional_term=2.0, constant_term=1.0)
        ),
    )

    assert isinstance(cost_curve.value_curve.function_data, LinearFunctionData)
    assert cost_curve.value_curve.function_data.proportional_term == 1.0
    assert isinstance(cost_curve.vom_cost.function_data, LinearFunctionData)
    assert cost_curve.vom_cost.function_data.proportional_term == 2.0


def test_fuel_curve():
    # Fuel curve
    fuel_curve = FuelCurve(
        value_curve=InputOutputCurve(
            function_data=LinearFunctionData(proportional_term=1.0, constant_term=2.0)
        ),
        vom_cost=InputOutputCurve(
            function_data=LinearFunctionData(proportional_term=2.0, constant_term=1.0)
        ),
        fuel_cost=2.5,
    )

    assert isinstance(fuel_curve.value_curve.function_data, LinearFunctionData)
    assert fuel_curve.value_curve.function_data.proportional_term == 1.0
    assert isinstance(fuel_curve.vom_cost.function_data, LinearFunctionData)
    assert fuel_curve.vom_cost.function_data.proportional_term == 2.0
    assert fuel_curve.fuel_cost == 2.5


def test_value_curve_custom_serialization():
    component = CurveComponent(
        name="test",
        cost_curve=CostCurve(
            value_curve=InputOutputCurve(
                function_data=LinearFunctionData(proportional_term=1.0, constant_term=2.0)
            ),
            vom_cost=InputOutputCurve(
                function_data=LinearFunctionData(proportional_term=2.0, constant_term=1.0)
            ),
        ),
    )

    model_dump = component.model_dump(mode="json")
    assert model_dump["cost_curve"]["value_curve"]["function_data"]["proportional_term"] == 1.0

    model_dump = component.model_dump(context={"magnitude_only": True})
    assert model_dump["cost_curve"]["value_curve"]["function_data"]["proportional_term"] == 1.0

    model_dump = component.model_dump(mode="json", context={"magnitude_only": True})
    assert model_dump["cost_curve"]["value_curve"]["function_data"]["proportional_term"] == 1.0


def test_value_curve_serialization(tmp_path):
    system = SimpleSystem(auto_add_composed_components=True)

    v1 = CurveComponent(
        name="test",
        cost_curve=CostCurve(
            value_curve=InputOutputCurve(
                function_data=LinearFunctionData(proportional_term=1.0, constant_term=2.0)
            ),
            vom_cost=InputOutputCurve(
                function_data=LinearFunctionData(proportional_term=2.0, constant_term=1.0)
            ),
        ),
    )
    system.add_component(v1)
    filename = tmp_path / "value_curve.json"

    system.to_json(filename, overwrite=True)
    system2 = SimpleSystem.from_json(filename)

    assert system2 is not None

    v2 = system2.get_component(CurveComponent, "test")

    assert v2 is not None
    assert isinstance(v1.cost_curve.value_curve.function_data, LinearFunctionData)
    assert (
        v1.cost_curve.value_curve.function_data.proportional_term
        == v2.cost_curve.value_curve.function_data.proportional_term
    )
    assert (
        v1.cost_curve.value_curve.function_data.constant_term
        == v2.cost_curve.value_curve.function_data.constant_term
    )



"""Test related to the pyarrow storage manager."""

import pytest
from datetime import datetime, timedelta
from pathlib import Path

import numpy as np
from loguru import logger

from infrasys.arrow_storage import ArrowTimeSeriesStorage
from infrasys.in_memory_time_series_storage import InMemoryTimeSeriesStorage
from infrasys.system import System
from infrasys.time_series_models import SingleTimeSeries

from .models.simple_system import SimpleSystem, SimpleBus, SimpleGenerator


@pytest.fixture(scope="session")
def test_system() -> System:
    system = SimpleSystem(time_series_directory=None)
    bus = SimpleBus(name="test-bus", voltage=1.1)
    gen1 = SimpleGenerator(name="gen1", active_power=1.0, rating=1.0, bus=bus, available=True)
    system.add_components(bus, gen1)
    return system


def test_file_creation(test_system: System):
    gen1 = test_system.get_component(SimpleGenerator, "gen1")
    ts = SingleTimeSeries.from_array(
        data=range(8784),
        variable_name="active_power",
        initial_time=datetime(year=2020, month=1, day=1),
        resolution=timedelta(hours=1),
    )
    test_system.time_series.add(ts, gen1, scenario="one", model_year="2030")
    time_series = test_system.time_series.get(gen1)
    assert isinstance(test_system.time_series.storage, ArrowTimeSeriesStorage)
    base_directory = test_system.get_time_series_directory()
    assert isinstance(base_directory, Path)
    time_series_fpath = base_directory.joinpath(str(time_series.uuid) + ".arrow")
    assert time_series_fpath.exists()


def test_copy_files(tmp_path):
    """Test that we can copy the time series from tmp to specified folder"""
    system = SimpleSystem()
    bus = SimpleBus(name="test-bus", voltage=1.1)
    gen1 = SimpleGenerator(name="gen1", active_power=1.0, rating=1.0, bus=bus, available=True)
    system.add_components(bus, gen1)
    ts = SingleTimeSeries.from_array(
        data=range(8784),
        variable_name="active_power",
        initial_time=datetime(year=2020, month=1, day=1),
        resolution=timedelta(hours=1),
    )
    system.time_series.add(ts, gen1, scenario="two", model_year="2030")
    filename = tmp_path / "system.json"
    system.to_json(filename)

    logger.info("Starting deserialization")
    system2 = SimpleSystem.from_json(filename)
    gen1b = system2.get_component(SimpleGenerator, gen1.name)
    time_series = system2.time_series.get(gen1b)
    time_series_fpath = (
        tmp_path / system2.get_time_series_directory() / (str(time_series.uuid) + ".arrow")
    )

    assert time_series_fpath.exists()


def test_read_deserialize_time_series(tmp_path):
    """Test that we can read from a deserialized system."""
    system = SimpleSystem()
    bus = SimpleBus(name="test-bus", voltage=1.1)
    gen1 = SimpleGenerator(name="gen1", active_power=1.0, rating=1.0, bus=bus, available=True)
    system.add_components(bus, gen1)
    ts = SingleTimeSeries.from_array(
        data=range(8784),
        variable_name="active_power",
        initial_time=datetime(year=2020, month=1, day=1),
        resolution=timedelta(hours=1),
    )
    system.time_series.add(ts, gen1, scenario="high", model_year="2030")
    filename = tmp_path / "system.json"
    system.to_json(filename)

    system2 = SimpleSystem.from_json(filename, time_series_directory=tmp_path)
    gen1b = system2.get_component(SimpleGenerator, gen1.name)
    deserialize_ts = system2.time_series.get(gen1b)
    assert isinstance(deserialize_ts, SingleTimeSeries)
    assert deserialize_ts.resolution == ts.resolution
    assert deserialize_ts.initial_time == ts.initial_time
    assert isinstance(deserialize_ts.data, np.ndarray)
    length = ts.length
    assert isinstance(length, int)
    assert np.array_equal(deserialize_ts.data, np.array(range(length)))


def test_copied_storage_system(simple_system_with_time_series):
    assert isinstance(
        simple_system_with_time_series._time_series_mgr._storage, ArrowTimeSeriesStorage
    )
    gen_component = next(simple_system_with_time_series.get_components(SimpleGenerator))
    data_array_1 = simple_system_with_time_series.list_time_series(gen_component)[0].data

    simple_system_with_time_series.convert_storage(time_series_in_memory=True, replace=True)

    assert isinstance(
        simple_system_with_time_series._time_series_mgr._storage, InMemoryTimeSeriesStorage
    )

    data_array_2 = simple_system_with_time_series.list_time_series(gen_component)[0].data
    assert np.array_equal(data_array_1, data_array_2)



import enum
import importlib
from typing import Any, Literal, Annotated, Type, Union
from uuid import UUID

from pydantic import Field, field_serializer

from infrasys.models import InfraSysBaseModel


TYPE_METADATA = "__metadata__"


class SerializedType(str, enum.Enum):
    """Controls how types are serialized."""

    BASE = "base"
    COMPOSED_COMPONENT = "composed_component"
    QUANTITY = "quantity"


class SerializedTypeBase(InfraSysBaseModel):
    """Applies to all normal types"""

    module: str
    type: str


class SerializedBaseType(SerializedTypeBase):
    """Applies to all normal types"""

    serialized_type: Literal[SerializedType.BASE] = SerializedType.BASE


class SerializedComponentReference(SerializedTypeBase):
    """Reference information for a component that has been serialized as a UUID within another."""

    serialized_type: Literal[SerializedType.COMPOSED_COMPONENT] = SerializedType.COMPOSED_COMPONENT
    uuid: UUID

    @field_serializer("uuid")
    def _serialize_uuid(self, _) -> str:
        return str(self.uuid)


class SerializedQuantityType(SerializedTypeBase):
    serialized_type: Literal[SerializedType.QUANTITY] = SerializedType.QUANTITY


class SerializedTypeMetadata(InfraSysBaseModel):
    """Serializes information about a type so that it can be de-serialized."""

    fields: Annotated[
        Union[
            SerializedBaseType,
            SerializedComponentReference,
            SerializedQuantityType,
        ],
        Field(discriminator="serialized_type"),
    ]


class CachedTypeHelper:
    """Helper class to deserialize types."""

    def __init__(self) -> None:
        self._observed_types: dict[tuple[str, str], Type] = {}
        self._deserialized_types: set[Type] = set()

    def add_deserialized_types(self, types: set[Type]) -> None:
        """Add types that have been deserialized."""
        self._deserialized_types.update(types)

    def allowed_to_deserialize(self, component_type: Type) -> bool:
        """Return True if the type can be deserialized."""
        return component_type in self._deserialized_types

    def get_type(self, metadata: SerializedTypeBase) -> Type:
        """Return the type contained in metadata, dynamically importing as necessary."""
        type_key = (metadata.module, metadata.type)
        component_type = self._observed_types.get(type_key)
        if component_type is None:
            component_type = _deserialize_type(*type_key)
            self._observed_types[type_key] = component_type
        return component_type


def serialize_value(obj: InfraSysBaseModel, *args, **kwargs) -> dict[str, Any]:
    """Serialize an infrasys object to a dictionary."""
    cls = type(obj)
    data = obj.model_dump(*args, mode="json", **kwargs)
    data[TYPE_METADATA] = SerializedTypeMetadata(
        fields=SerializedBaseType(
            module=cls.__module__,
            type=cls.__name__,
        ),
    ).model_dump()
    return data


def deserialize_type(metadata: SerializedTypeBase) -> Type:
    """Dynamically import the type and return it."""
    return _deserialize_type(metadata.module, metadata.type)


def _deserialize_type(module, obj_type) -> Type:
    mod = importlib.import_module(module)
    return getattr(mod, obj_type)


def deserialize_value(data: dict[str, Any], metadata: SerializedTypeBase) -> Any:
    """Deserialize the value from a dictionary."""
    ctype = deserialize_type(metadata)
    return ctype(**data)



"""This module defines basic unit quantities.

To create new Quantities for a given base unit, we just need to specify the
base unit as the second argument of `ureg.check`.
"""

from infrasys.base_quantity import BaseQuantity

# ruff:noqa
# fmt: off

class Distance(BaseQuantity): __base_unit__ = "meter"

class Voltage(BaseQuantity): __base_unit__ = "volt"

class Current(BaseQuantity): __base_unit__ = "ampere"

class Angle(BaseQuantity): __base_unit__ = "degree"

class ActivePower(BaseQuantity): __base_unit__ = "watt"

class Energy(BaseQuantity): __base_unit__ = "watthour"

class Time(BaseQuantity): __base_unit__ = "minute"

class Resistance(BaseQuantity): __base_unit__ = "ohm"



from .models.simple_system import SimpleSystem, SimpleBus, SimpleGenerator
from infrasys.time_series_models import SingleTimeSeries
from infrasys.exceptions import ISAlreadyAttached
from infrasys.arrow_storage import ArrowTimeSeriesStorage
from infrasys.in_memory_time_series_storage import InMemoryTimeSeriesStorage
from datetime import timedelta, datetime
import numpy as np
import pytest


def get_data_and_uuids(system):
    uuids = system._time_series_mgr.metadata_store.unique_uuids_by_type(SingleTimeSeries.__name__)
    data = {
        uuid: system._time_series_mgr._storage.get_raw_single_time_series(uuid) for uuid in uuids
    }
    return uuids, data


@pytest.mark.parametrize(
    "original_kwargs,new_kwargs,original_stype,new_stype",
    [
        ({"time_series_in_memory": True}, {}, InMemoryTimeSeriesStorage, ArrowTimeSeriesStorage),
        ({}, {"time_series_in_memory": True}, ArrowTimeSeriesStorage, InMemoryTimeSeriesStorage),
    ],
)
def test_memory_convert_storage_time_series(
    original_kwargs, new_kwargs, original_stype, new_stype
):
    test_bus = SimpleBus.example()
    test_generator = SimpleGenerator.example()
    system = SimpleSystem(auto_add_composed_components=True, **original_kwargs)

    assert isinstance(system._time_series_mgr._storage, original_stype)

    system.get_components()
    system.add_components(test_bus)
    system.add_components(test_generator)

    test_time_series_data = SingleTimeSeries(
        data=np.arange(24),
        resolution=timedelta(hours=1),
        initial_time=datetime.now(),
        variable_name="load",
    )
    system.add_time_series(test_time_series_data, test_generator)
    with pytest.raises(ISAlreadyAttached):
        system.add_time_series(test_time_series_data, test_generator)

    original_uuids, original_data = get_data_and_uuids(system)

    system.convert_storage(**new_kwargs)

    assert isinstance(system._time_series_mgr._storage, new_stype)
    new_uuids, new_data = get_data_and_uuids(system)

    assert set(original_uuids) == set(new_uuids)

    for uuid in new_uuids:
        assert np.array_equal(original_data[uuid], new_data[uuid])



"""Normalization utilities"""

import abc
from enum import Enum
from typing import Literal, Optional, Annotated, Union

import numpy as np
from numpy.typing import NDArray
from pydantic import Field

from infrasys.models import InfraSysBaseModel


class NormalizationType(str, Enum):
    MAX = "max"
    BY_VALUE = "by_value"


class NormalizationBase(InfraSysBaseModel, abc.ABC):
    """Base class for all normalization models"""

    @abc.abstractmethod
    def normalize_array(self, data: np.ndarray) -> np.ndarray:
        """Normalize the array."""


class NormalizationMax(NormalizationBase):
    """Perform normalization by the max value in an array."""

    max_value: Optional[float] = None
    normalization_type: Literal[NormalizationType.MAX] = NormalizationType.MAX

    def normalize_array(self, data: NDArray) -> NDArray:
        self.max_value = np.max(data)
        return data / self.max_value


class NormalizationByValue(NormalizationBase):
    """Perform normalization by a user-defined value."""

    value: float
    normalization_type: Literal[NormalizationType.BY_VALUE] = NormalizationType.BY_VALUE

    def normalize_array(self, data: NDArray) -> NDArray:
        return data / self.value


NormalizationModel = Annotated[
    Union[None, NormalizationMax, NormalizationByValue],
    Field(
        description="Defines the type of normalization performed on the data, if any.",
        discriminator="normalization_type",
    ),
]
